export fit_sellmeier

const GUESS_MARGIN = 0.05  # how much λres guesses are separated from range of measured λ

function has_similar(λres::AbsVecReal; rtol=1e-3)  # assume λres is sorted
    for i in eachindex(λres)[begin:end-1]
        isapprox(λres[i], λres[i+1]; rtol) && return true  # if poles are different by less than 0.1%, return true
    end

    return false
end

"""
    fit_sellmeier(filepath::String[, args...]; kwargs...)

Read `λ` and `ε` from the file at `filepath` and call `fit_sellmeier(λ, ε[, args...]; kwargs...)`.
"""
fit_sellmeier(filepath::String, args...; kwargs...) = fit_sellmeier(read(filepath)..., args...; kwargs...)

"""
    fit_sellmeier(λ::AbstractVector{<:Real}, ε::AbstractVector{<:Real}; λmin::Real=-Inf, λmax::Real=Inf, Nmax::Integer=10, rtol_λres::Real=1e-3)

Fit the dielectric constant data `ε` sampled at wavelengths `λ` to the Sellmeier equation
with up to `Nmax` terms.  Only the entries of `λ` between `λmin` and `λmax` and the
corresponding entries of `ε` are used.  If `λres` of different terms have relative
difference less than`rtol_λres`, they are considered the same term and merged.
"""
fit_sellmeier(λ::AbsVecReal, ε::AbsVecReal; kwargs...) = fit_sellmeier(float(λ), float(ε); kwargs...)
function fit_sellmeier(λ::AbsVecFloat,  # wavelengths where ε was measured
                       ε::AbsVecFloat;  # measured relative permittivities (= squared refractive indices)
                       λmin::Real=-Inf,  # lower bound of λ entries to use
                       λmax::Real=Inf,  # upper bound of λ entries to use
                       Nmax::Integer=10,  # maximum number of terms in Sellmeier equation
                       rtol_λres::Real=1e-3)  # relative tolerance to distinguish λres
    me = fit_sellmeier(λ, ε, Val(1); λmin, λmax)  # me: model and error
    for N = 2:Nmax
        me′ = fit_sellmeier(λ, ε, Val(N); λmin, λmax)
        if me′.err > me.err  # if more terms than necessary are used, error starts to increase
            break
        elseif has_similar(me′.mdl.λres; rtol=rtol_λres)  # terms with similar poles are unnecessary
            break
        else
            me = me′
        end
    end

    return me
end

"""
    fit_sellmeier(λ::AbstractVector{<:Real}, ε::AbstractVector{<:Real}, N::Integer; λmin::Real=-Inf, λmax::Real=Inf)

Fit the dielectric constant data `ε` sampled at wavelengths `λ` to the Sellmeier equation
with exactly `N` terms.  Only the entries of `λ` between `λmin` and `λmax` and the
corresponding entries of `ε` are used.  Among the `N` terms, some are below and the other
are above the range of `λ`.
"""
fit_sellmeier(λ::AbsVecReal, ε::AbsVecReal, N::Integer; kwargs...) = fit_sellmeier(λ, ε, Val(N); kwargs...)
fit_sellmeier(λ::AbsVecReal, ε::AbsVecReal, ::Val{N}; kwargs...) where {N} = fit_sellmeier(float(λ), float(ε), Val(N); kwargs...)
function fit_sellmeier(λ::AbsVecFloat,  # wavelengths where ε was measured
                       ε::AbsVecFloat,  # measured relative permittivities (= squared refractive indices)
                       ::Val{N};  # number of terms in Sellmeier model
                       λmin::Real=-Inf,  # lower bound of λ entries to use
                       λmax::Real=Inf  # upper bound of λ entries to use
                       ) where {N}
    local me
    err_min = Inf
    for Nₙ = 0:N
        Nₚ = N - Nₙ
        me′ = fit_sellmeier(λ, ε, Nₙ, Nₚ; λmin, λmax)
        if me′.err < err_min
            err_min = me′.err
            me = me′
        end
    end

    return me
end

"""
    fit_sellmeier(λ::AbstractVector{<:Real}, ε::AbstractVector{<:Real}, Nₙ::Integer, Nₚ::Integer; λmin::Real=-Inf, λmax::Real=Inf)

Fit the dielectric constant data `ε` sampled at wavelengths `λ` to the Sellmeier equation
with exactly `Nₙ` terms below and `Nₚ` terms above the range of `λ`.  Only the entries of `λ`
between `λmin` and `λmax` and the corresponding entries of `ε` are used.
"""
fit_sellmeier(λ::AbsVecReal, ε::AbsVecReal, Nₙ::Integer, Nₚ::Integer; kwargs...) = fit_sellmeier(float(λ), float(ε), Nₙ, Nₚ; kwargs...)
function fit_sellmeier(λ::AbsVecFloat,
                       ε::AbsVecFloat,
                       Nₙ::Integer,  # number of terms with poles below λ range
                       Nₚ::Integer;  # number of terms with poles above λ range
                       λmin::Real=-Inf,  # lower bound of λ entries to use
                       λmax::Real=Inf)  # upper bound of λ entries to use
    issorted(λ) || @error "λ = $λ should be sorted."

    N = Nₙ + Nₚ

    # Take the data in the valid range.
    ind_first = findfirst(λᵢ->λᵢ≥λmin, λ)
    ind_last = findlast(λᵢ->λᵢ≤λmax, λ)
    λ = @view λ[ind_first:ind_last]
    ε = @view ε[ind_first:ind_last]

    # Normaize λ.
    λ_first, λ_last = λ[1], λ[end]
    λ₀ = √(λ_first * λ_last)
    λ̂ = λ ./ λ₀
    Nλ = length(λ̂)

    # Create guess parameters.
    λ̂ₙ, λ̂ₚ = (1-GUESS_MARGIN)λ̂[1], (1+GUESS_MARGIN)λ̂[end]

    p₀mat = ones(N,2)
    p₀mat[1:Nₙ,2] .= λ̂ₙ
    p₀mat[Nₙ+1:end,2] .= λ̂ₚ
    p₀ = reshape(p₀mat, :)

    # Perform fitting.
    fit = curve_fit((λ̂,p)->(BC = reshape(p,:,2);
                            str = Ref(@view(BC[:,1]));
                            λ̂res = Ref(@view(BC[:,2]));
                            sellmeier_model.(λ̂, str, λ̂res)),
                    λ̂, ε, p₀)

    str = @view(fit.param[1:N])  # entries of str are dimensionless, so no need to unnormalize them
    λ̂res = @view(fit.param[N+1:2N])
    λres = λ̂res .* λ₀  # entries of λres are wavelengths, so unnormalize them
    err = sqrt(sum(abs2, fit.resid) / Nλ)  # RMS error

    mdl = SellmeierModel(str, λres)

    return (; mdl, err)
end
