export RefractiveIndexInfo

"""
    SpectralVariable

Used to specify the type of the spectral variable which the refractive index is a function
of.  The available instances are `WAVELENGTH` for wavelength, `FREQUENCY` for frequency, and
`ENERGY` for photon energy.
"""
@enum SpectralVariable WAVELENGTH FREQUENCY ENERGY
for ins in instances(SpectralVariable); @eval export $(Symbol(ins)); end  # export all instances

"""
    read(filepath::String; kwargs...)

Read the refractive index file at `filepath`, and return the wavelength vector `λ` and the
dielectric constant vector `ε` as a named tuple.

# Keyword arguments
- `s_var::SpectralVariable=WAVELENGTH`: `WAVELENGTH`, `FREQUENCY`, or `ENERGY` if the refractive index file describes refractive index as a function of wavelength, frequency, or photon energy.
- `unit_prefix::Real=micro`: the prefix to multiply to the values of the spectral variable to convert their units to the SI units.  For example, `1e-6` for wavelengths in units of µm; `1.602176643` for photon energies in units of eV.
- `s_col::Integer=1`: the index of the spectral variable column in the refractive index file.
- `n_col::Integer=2`: the index of the refractive index column in the refractive index file.
- `kwargs...`: the keyword arguments to pass to `CSV.read()`
"""
function read(filepath::String;  # location of CSV file containing λ, n, and possibly k columns
              s_var::SpectralVariable=WAVELENGTH,
              unit_prefix::Real=micro,  # e.g., micro for μm; tera for THz; e⁺ for eV
              s_col::Integer=1,  # column index of spectral variable
              n_col::Integer=2,  # column index of refractive index
              n_tbl::Integer=1,  # index of table to use, in case of multiple tables in single CSV file
              kwargs...)  # keyword arguments for CSV.read; can be used to read .txt
    λ = Float[]
    ε = Float[]

    rows = CSV.Rows(filepath; types=String, kwargs...)
    i_tbl = 1
    for row in rows
        num_col = length(row)
        num_col≥2 || @error "Dataset at $filepath should contain at least two columns corresponding to spectral variable λ and refractive index n, but it contains $num_col column."

        if all(isnothing.(tryparse.(Float,row)))
            i_tbl += 1
            i_tbl==n_tbl+1 && break
        end

        i_tbl<n_tbl && continue

        s = parse(Float,row[s_col]) * unit_prefix
        n = parse(Float,row[n_col])

        if s_var == WAVELENGTH
            push!(λ, s)
        elseif s_var == FREQUENCY  # s = ν
            push!(λ, c₀/s)  # λ = c₀ / ν
        else  # s_var == ENERGY; s = E
            push!(λ, ℎc₀/s)  # λ = c₀ / ν = hc₀ / hν = ℎc₀ / E
        end

        push!(ε, n^2)
    end

    # If sorted in descending order, reverse it.
    issorted(λ) || issorted(λ, rev=true) || @error "λ = $λ should be sorted in ascending or descending order."
    if issorted(λ, rev=true)  # sorted in descending order
        reverse!(λ)
        reverse!(ε)
    end

    return (; λ, ε)
end

const URL_RI_INFO = "https://RefractiveIndex.Info/tmp/data/main"
const DIR_RI_INFO = pkgdir(@__MODULE__, "RefractiveIndex.INFO")

"""
    RefractiveIndexInfo(material::String, author::String; kwargs...)

Download the CSV file containing the refractive index vs. wavelength data from the
RefractiveIndex.Info website, and return the local path to the downloaded file.  By default,
the file is named `<first author name>.csv`` and stored at `<SellmeierFit package directory>/<material name>/`.

`<first author name>` and `<material name>` correspond to the dataset-distinguishing portion
of the URL of the RefractiveIndex.Info webpage.  For example, the webpage for the silicon
dioxide dataset published by Malitson et al. is at https://refractiveindex.info/?shelf=main&book=SiO2&page=Malitson,
where `SiO2` is `<material name>` and `Malitson` is `<first author name>`.

# Keyword arguments
- `dir::String=""`: directory name where the downloaded file is stored.
- `file::String=""`: file name to which the downloaded file is named.
- `force_download::Bool=false`: true to force download; false to avoid downloading when the file already exists
- `max_att::Integer=3`: the maximum number of download attempts.
- `verbose::Bool=true`: true to display messages; false to silence them.
"""
function RefractiveIndexInfo(material::String,
                             author::String;
                             dir::String="",
                             file::String="",
                             force_download::Bool=false,
                             max_att::Integer=3,
                             verbose::Bool=true)
    isempty(dir) && (dir = joinpath(DIR_RI_INFO, material))
    isempty(file) && (file = author * ".csv")

    filepath = joinpath(dir, file)
    if !force_download && isfile(filepath)
        verbose && println("Using already existing $filepath.")
    else
        url = join((URL_RI_INFO, material, author * ".csv"), "/")

        local tempfile
        for num_att = 1:max_att
            try
                print("Attempt $num_att to download $url: ")
                tempfile = Downloads.download(url)
            catch e
                if num_att < max_att
                    println("Fails.")
                    continue
                else  # num_att == max_att
                    println("Fails; Abort.")
                    throw(e)
                end
            end

            break
        end

        println("Succeeds.")
        !isdir(dir) && mkpath(dir)
        mv(tempfile, filepath)
    end

    return filepath
end
